#! /usr/bin/python
#
#    Copyright 2002 Jeremy Avnet <brainsik-gpl@theory.org>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    A copy of the license can be found at http://www.gnu.org/licenses/gpl.txt
#

import gd
import string

from math   import exp
from random import randrange, random


class Magnet:

  def __init__(self, size, T=1, p=0.5, world='torus', nbrhd='5-cell'):
    self.size     = size   # length of world edge
    self.world    = world  # world type
    self.nbrhd    = nbrhd  # neihborhood type

    self.img      = gd.image((size,size))                  # lattice graphic
    self.colorneg = self.img.colorAllocate((64,64,64))     # (negative color)
    self.colorpos = self.img.colorAllocate((172,172,255))  # (positive color)

    self.T        = T     # temperature
    self.M        = 0     # total magnetization
    self.H        = 0     # total energy
    
    self.initLattice(p)   # a new lattice of atoms
    self.initImage()      # the lattice graphic
  
    self.calcM()          # total magnetization
    self.calcH()          # total energy


  # creates lattice of up spins with probability p (down spins 1 - p)
  def initLattice(self, p):
    lattice = {}
    for y in range(self.size):
      for x in range(self.size):
        if (random() < p):
          lattice[(x,y)] = 1
        else:
          lattice[(x,y)] = -1
    self.lattice = lattice


  # creates a graphic based on current lattice
  def initImage(self):
    for y in range(self.size):
      for x in range(self.size):
        if (self.lattice[(x,y)] + 1):
          self.img.setPixel((x,y),self.colorpos)
        else:
          self.img.setPixel((x,y),self.colorneg)


  # calculates H for a particular atom in the lattice
  def Hatom(self,x,y):
    if self.world == 'torus':
      return (-1 * (self.lattice[(x,y)] * self.lattice[((x + 1) % self.size, y)] +
                    self.lattice[(x,y)] * self.lattice[((x - 1 + self.size) % self.size, y)] +
                    self.lattice[(x,y)] * self.lattice[(x, (y + 1) % self.size)] +
                    self.lattice[(x,y)] * self.lattice[(x, (y - 1 + self.size) % self.size)]))

  
  # calulates H(lattice)
  def calcH(self):
    energy = 0
    for y in range(self.size):
      for x in range(self.size):
        energy += self.Hatom(x,y)
    self.H = energy


  # calculates M(lattice)
  def calcM(self):
    magnetization = 0
    for y in range(self.size):
      for x in range(self.size):
        magnetization += self.lattice[(x,y)]
    self.M = magnetization    
    

  # Monte Carlo simulation of the system
  def iterate(self, its=1):
    for i in range(its):
      # choose random atom
      x = randrange(self.size)
      y = randrange(self.size)

      deltaH  = (-2) * self.Hatom(x,y)
      
#      print "------------------------------"
#      print "(x, y) =", (x, y)
#      print "delta H =", deltaH

      if (deltaH <= 0):
        self.lattice[(x,y)] *= -1
        self.M += 2 * self.lattice[(x,y)]
        self.H += deltaH
        if (self.lattice[(x,y)] + 1):
          self.img.setPixel((x,y),self.colorpos)
        else:
          self.img.setPixel((x,y),self.colorneg)
      else:
        P = exp(-1 * deltaH / self.T)
        if (random() < P):
          self.lattice[(x,y)] *= -1
          self.M += 2 * self.lattice[(x,y)]
          self.H += deltaH
          if (self.lattice[(x,y)] + 1):
            self.img.setPixel((x,y),self.colorpos)
          else:
            self.img.setPixel((x,y),self.colorneg)


  def writeImage(self, name="magnet"):
    
    f = open(name + ".png", "w")
    self.img.writePng(f)
    f.close


size = 200
temp = 0.001
prob = 0.5
type = 'torus'

M = Magnet(size, temp, prob, type)
#M.writeImage("000")
for x in range(50):
  M.iterate(10000)
  M.writeImage(string.zfill(x,3))

